@@ -1,3 +1,5 @@
+from __future__ import annotations
+
 import abc
 import ipaddress
 import json
@@ -7,35 +9,40 @@ import threading
 import time
 import urllib.parse
 from collections import defaultdict
+from collections.abc import Iterable
+from collections.abc import Mapping
+from collections.abc import MutableMapping
 from contextlib import contextmanager
 from contextlib import suppress
 from copy import copy
 from enum import Enum
-from ssl import SSLContext
+from re import Pattern
+from typing import TYPE_CHECKING
 from typing import Any
 from typing import Callable
-from typing import Iterable
-from typing import List
-from typing import Mapping
-from typing import MutableMapping
+from typing import ClassVar
 from typing import Optional
-from typing import Pattern
-from typing import Tuple
 from typing import Union
 
 import werkzeug.http
+from werkzeug import Request
+from werkzeug import Response
 from werkzeug.datastructures import Authorization
 from werkzeug.datastructures import MultiDict
 from werkzeug.serving import make_server
-from werkzeug.wrappers import Request
-from werkzeug.wrappers import Response
+
+if TYPE_CHECKING:
+    from ssl import SSLContext
+    from types import TracebackType
+
+    from werkzeug.serving import BaseWSGIServer
 
 URI_DEFAULT = ""
 METHOD_ALL = "__ALL"
 
 HEADERS_T = Union[
     Mapping[str, Union[str, Iterable[str]]],
-    Iterable[Tuple[str, str]],
+    Iterable[tuple[str, str]],
 ]
 
 HVMATCHER_T = Callable[[str, Optional[str], str], bool]
@@ -54,32 +61,24 @@ class Error(Exception):
     Base class for all exception defined in this package.
     """
 
-    pass
-
 
 class NoHandlerError(Error):
     """
     Raised when a :py:class:`RequestHandler` has no registered method to serve the request.
     """
 
-    pass
-
 
 class HTTPServerError(Error):
     """
     Raised when there's a problem with HTTP server.
     """
 
-    pass
-
 
 class NoMethodFoundForMatchingHeaderValueError(Error):
     """
     Raised when a :py:class:`HeaderValueMatcher` has no registered method to match the header value.
     """
 
-    pass
-
 
 class WaitingSettings:
     """Class for providing default settings and storing them in HTTPServer
@@ -89,7 +88,12 @@ class WaitingSettings:
     :param timeout: time (in seconds) until time is out
     """
 
-    def __init__(self, raise_assertions: bool = True, stop_on_nohandler: bool = True, timeout: float = 5):
+    def __init__(
+        self,
+        raise_assertions: bool = True,  # noqa: FBT001
+        stop_on_nohandler: bool = True,  # noqa: FBT001
+        timeout: float = 5,
+    ):
         self.raise_assertions = raise_assertions
         self.stop_on_nohandler = stop_on_nohandler
         self.timeout = timeout
@@ -105,17 +109,19 @@ class Waiting:
         self._start = time.monotonic()
         self._stop = None
 
-    def complete(self, result: bool):
+    def complete(self, result: bool):  # noqa: FBT001
         self._result = result
         self._stop = time.monotonic()
 
     @property
     def result(self) -> bool:
-        return self._result
+        return bool(self._result)
 
     @property
     def elapsed_time(self) -> float:
         """Elapsed time in seconds"""
+        if self._stop is None:
+            raise TypeError("unsupported operand type(s) for -: 'NoneType' and 'float'")
         return self._stop - self._start
 
 
@@ -127,23 +133,23 @@ class HeaderValueMatcher:
         and return whether they are equal as bool.
     """
 
-    DEFAULT_MATCHERS: MutableMapping[str, Callable[[Optional[str], str], bool]] = {}
+    DEFAULT_MATCHERS: ClassVar[MutableMapping[str, Callable[[str | None, str], bool]]] = {}
 
-    def __init__(self, matchers: Optional[Mapping[str, Callable[[Optional[str], str], bool]]] = None):
+    def __init__(self, matchers: Mapping[str, Callable[[str | None, str], bool]] | None = None):
         self.matchers = self.DEFAULT_MATCHERS if matchers is None else matchers
 
     @staticmethod
-    def authorization_header_value_matcher(actual: Optional[str], expected: str) -> bool:
-        callable = getattr(Authorization, "from_header", None)
-        if callable is None:  # Werkzeug < 2.3.0
-            callable = werkzeug.http.parse_authorization_header
-        return callable(actual) == callable(expected)
+    def authorization_header_value_matcher(actual: str | None, expected: str) -> bool:
+        func = getattr(Authorization, "from_header", None)
+        if func is None:  # Werkzeug < 2.3.0
+            func = werkzeug.http.parse_authorization_header  # type: ignore[attr-defined]
+        return func(actual) == func(expected)  # type: ignore
 
     @staticmethod
-    def default_header_value_matcher(actual: Optional[str], expected: str) -> bool:
+    def default_header_value_matcher(actual: str | None, expected: str) -> bool:
         return actual == expected
 
-    def __call__(self, header_name: str, actual: Optional[str], expected: str) -> bool:
+    def __call__(self, header_name: str, actual: str | None, expected: str) -> bool:
         try:
             matcher = self.matchers[header_name]
         except KeyError:
@@ -172,7 +178,7 @@ class QueryMatcher(abc.ABC):
         return values[0] == values[1]
 
     @abc.abstractmethod
-    def get_comparing_values(self, request_query_string: bytes) -> tuple:
+    def get_comparing_values(self, request_query_string: bytes) -> tuple[Any, Any]:
         pass
 
 
@@ -181,7 +187,7 @@ class StringQueryMatcher(QueryMatcher):
     Matches a query for a string or bytes specified
     """
 
-    def __init__(self, query_string: Union[bytes, str]):
+    def __init__(self, query_string: bytes | str):
         """
         :param query_string: the query string will be compared to this string or bytes.
             If string is specified, it will be encoded by the encode() method.
@@ -193,10 +199,10 @@ class StringQueryMatcher(QueryMatcher):
 
         self.query_string = query_string
 
-    def get_comparing_values(self, request_query_string: bytes) -> tuple:
+    def get_comparing_values(self, request_query_string: bytes) -> tuple[bytes, bytes]:
         if isinstance(self.query_string, str):
             query_string = self.query_string.encode()
-        elif isinstance(self.query_string, bytes):
+        elif isinstance(self.query_string, bytes):  # type: ignore
             query_string = self.query_string
         else:
             raise TypeError("query_string must be a string, or a bytes-like object")
@@ -209,7 +215,7 @@ class MappingQueryMatcher(QueryMatcher):
     Matches a query string to a dictionary or MultiDict specified
     """
 
-    def __init__(self, query_dict: Union[Mapping, MultiDict]):
+    def __init__(self, query_dict: Mapping[str, str] | MultiDict[str, str]):
         """
         :param query_dict: if dictionary (Mapping) is specified, it will be used as a
             key-value mapping where both key and value should be string. If there are multiple
@@ -219,7 +225,7 @@ class MappingQueryMatcher(QueryMatcher):
         """
         self.query_dict = query_dict
 
-    def get_comparing_values(self, request_query_string: bytes) -> tuple:
+    def get_comparing_values(self, request_query_string: bytes) -> tuple[Mapping[str, str], Mapping[str, str]]:
         query = MultiDict(urllib.parse.parse_qsl(request_query_string.decode("utf-8")))
         if isinstance(self.query_dict, MultiDict):
             return (query, self.query_dict)
@@ -232,26 +238,26 @@ class BooleanQueryMatcher(QueryMatcher):
     Matches the query depending on the boolean value
     """
 
-    def __init__(self, result: bool):
+    def __init__(self, result: bool):  # noqa: FBT001
         """
         :param result: if this parameter is true, the query match will be always
             successful. Otherwise, no query match will be successful.
         """
         self.result = result
 
-    def get_comparing_values(self, request_query_string):
+    def get_comparing_values(self, request_query_string: bytes):  # noqa: ARG002
         if self.result:
             return (True, True)
         else:
             return (True, False)
 
 
-def _create_query_matcher(query_string: Union[None, QueryMatcher, str, bytes, Mapping]) -> QueryMatcher:
+def _create_query_matcher(query_string: None | QueryMatcher | str | bytes | Mapping[str, str]) -> QueryMatcher:
     if isinstance(query_string, QueryMatcher):
         return query_string
 
     if query_string is None:
-        return BooleanQueryMatcher(True)
+        return BooleanQueryMatcher(result=True)
 
     if isinstance(query_string, (str, bytes)):
         return StringQueryMatcher(query_string)
@@ -272,7 +278,6 @@ class URIPattern(abc.ABC):
             with "/" and does not contain the query part.
         :return: True if there's a match, False otherwise
         """
-        pass
 
 
 class RequestMatcher:
@@ -306,13 +311,13 @@ class RequestMatcher:
 
     def __init__(
         self,
-        uri: Union[str, URIPattern, Pattern[str]],
+        uri: str | URIPattern | Pattern[str],
         method: str = METHOD_ALL,
-        data: Union[str, bytes, None] = None,
+        data: str | bytes | None = None,
         data_encoding: str = "utf-8",
-        headers: Optional[Mapping[str, str]] = None,
-        query_string: Union[None, QueryMatcher, str, bytes, Mapping] = None,
-        header_value_matcher: Optional[HVMATCHER_T] = None,
+        headers: Mapping[str, str] | None = None,
+        query_string: None | QueryMatcher | str | bytes | Mapping[str, str] = None,
+        header_value_matcher: HVMATCHER_T | None = None,
         json: Any = UNDEFINED,
     ):
         if json is not UNDEFINED and data is not None:
@@ -383,7 +388,7 @@ class RequestMatcher:
             #
             # also, python will raise TypeError when self.uri is a conflicting type
 
-            return self.uri == URI_DEFAULT or path == self.uri
+            return self.uri in (URI_DEFAULT, path)
 
     def match_json(self, request: Request) -> bool:
         """
@@ -409,7 +414,7 @@ class RequestMatcher:
 
         return json_received == self.json
 
-    def difference(self, request: Request) -> List[Tuple]:
+    def difference(self, request: Request) -> list[tuple[str, str, str | URIPattern]]:
         """
         Calculates the difference between the matcher and the request.
 
@@ -421,19 +426,19 @@ class RequestMatcher:
         matches the fields set in the matcher object.
         """
 
-        retval: List[Tuple] = []
+        retval: list[tuple[str, Any, Any]] = []
 
         if not self.match_uri(request):
             retval.append(("uri", request.path, self.uri))
 
-        if self.method != METHOD_ALL and self.method != request.method:
+        if self.method not in (METHOD_ALL, request.method):
             retval.append(("method", request.method, self.method))
 
         if not self.query_matcher.match(request.query_string):
             retval.append(("query_string", request.query_string, self.query_string))
 
-        request_headers = {}
-        expected_headers = {}
+        request_headers: dict[str, str | None] = {}
+        expected_headers: dict[str, str] = {}
         for key, value in self.headers.items():
             if not self.header_value_matcher(key, request.headers.get(key), value):
                 request_headers[key] = request.headers.get(key)
@@ -466,9 +471,9 @@ class RequestHandlerBase(abc.ABC):
 
     def respond_with_json(
         self,
-        response_json,
+        response_json: Any,
         status: int = 200,
-        headers: Optional[Mapping[str, str]] = None,
+        headers: Mapping[str, str] | None = None,
         content_type: str = "application/json",
     ):
         """
@@ -485,16 +490,16 @@ class RequestHandlerBase(abc.ABC):
 
     def respond_with_data(
         self,
-        response_data: Union[str, bytes] = "",
+        response_data: str | bytes = "",
         status: int = 200,
-        headers: Optional[HEADERS_T] = None,
-        mimetype: Optional[str] = None,
-        content_type: Optional[str] = None,
+        headers: HEADERS_T | None = None,
+        mimetype: str | None = None,
+        content_type: str | None = None,
     ):
         """
         Prepares a response with raw data.
 
-        For detailed description please see the :py:class:`werkzeug.wrappers.Response` object as the
+        For detailed description please see the :py:class:`werkzeug.Response` object as the
         parameters are analogue.
 
         :param response_data: a string or bytes object representing the body of the response
@@ -514,8 +519,6 @@ class RequestHandlerBase(abc.ABC):
         :param response: the response object which will be responded
         """
 
-        pass
-
 
 class RequestHandler(RequestHandlerBase):
     """
@@ -529,7 +532,12 @@ class RequestHandler(RequestHandlerBase):
 
     def __init__(self, matcher: RequestMatcher):
         self.matcher = matcher
-        self.request_handler: Optional[Callable[[Request], Response]] = None
+        self.request_handler: Callable[[Request], Response] | None = None
+        self._hooks: list[Callable[[Request, Response], Response]] = []
+
+    def with_post_hook(self, hook: Callable[[Request, Response], Response]):
+        self._hooks.append(hook)
+        return self
 
     def respond(self, request: Request) -> Response:
         """
@@ -547,7 +555,11 @@ class RequestHandler(RequestHandlerBase):
                 "Matching request handler found but no response defined: {} {}".format(request.method, request.path)
             )
         else:
-            return self.request_handler(request)
+            response = self.request_handler(request)
+
+            for hook in self._hooks:
+                response = hook(request, response)
+            return response
 
     def respond_with_handler(self, func: Callable[[Request], Response]):
         """
@@ -560,14 +572,23 @@ class RequestHandler(RequestHandlerBase):
     def respond_with_response(self, response: Response):
         self.request_handler = lambda request: response
 
+    def __repr__(self) -> str:
+        class_name = self.__class__.__name__
+        retval = (
+            f"<{class_name} uri={self.matcher.uri!r} method={self.matcher.method!r} "
+            f"query_string={self.matcher.query_string!r} headers={self.matcher.headers!r} data={self.matcher.data!r} "
+            f"json={self.matcher.json!r}>"
+        )
+        return retval
 
-class RequestHandlerList(list):
+
+class RequestHandlerList(list[RequestHandler]):
     """
     Represents a list of :py:class:`RequestHandler` objects.
 
     """
 
-    def match(self, request: Request) -> Optional[RequestHandler]:
+    def match(self, request: Request) -> RequestHandler | None:
         """
         Returns the first request handler which matches the specified request. Otherwise, it returns `None`.
         """
@@ -590,11 +611,12 @@ class HTTPServerBase(abc.ABC):  # pylint: disable=too-many-instance-attributes
     :param host: the host or IP where the server will listen
     :param port: the TCP port where the server will listen
     :param ssl_context: the ssl context object to use for https connections
+    :param threaded: whether to handle concurrent requests in separate threads
 
     .. py:attribute:: log
 
         Attribute containing the list of two-element tuples. Each tuple contains
-        :py:class:`werkzeug.wrappers.Request` and :py:class:`werkzeug.wrappers.Response` object which represents the
+        :py:class:`werkzeug.Request` and :py:class:`werkzeug.Response` object which represents the
         incoming request and the outgoing response which happened during the lifetime
         of the server.
 
@@ -610,7 +632,9 @@ class HTTPServerBase(abc.ABC):  # pylint: disable=too-many-instance-attributes
         self,
         host: str,
         port: int,
-        ssl_context: Optional[SSLContext] = None,
+        ssl_context: SSLContext | None = None,
+        *,
+        threaded: bool = False,
     ):
         """
         Initializes the instance.
@@ -618,12 +642,13 @@ class HTTPServerBase(abc.ABC):  # pylint: disable=too-many-instance-attributes
         """
         self.host = host
         self.port = port
-        self.server = None
-        self.server_thread = None
-        self.assertions: List[str] = []
-        self.handler_errors: List[Exception] = []
-        self.log: List[Tuple[Request, Response]] = []
+        self.server: BaseWSGIServer | None = None
+        self.server_thread: threading.Thread | None = None
+        self.assertions: list[str | AssertionError] = []
+        self.handler_errors: list[Exception] = []
+        self.log: list[tuple[Request, Response]] = []
         self.ssl_context = ssl_context
+        self.threaded = threaded
         self.no_handler_status_code = 500
 
     def __repr__(self):
@@ -706,7 +731,7 @@ class HTTPServerBase(abc.ABC):  # pylint: disable=too-many-instance-attributes
 
         This should not be called directly, but can be overridden to tailor it to your needs.
         """
-
+        assert self.server is not None
         self.server.serve_forever()
 
     def is_running(self) -> bool:
@@ -715,18 +740,18 @@ class HTTPServerBase(abc.ABC):  # pylint: disable=too-many-instance-attributes
         """
         return bool(self.server)
 
-    def start(self):
+    def start(self) -> None:
         """
         Start the server in a thread.
 
         This method returns immediately (e.g. does not block), and it's the caller's
         responsibility to stop the server (by calling :py:meth:`stop`) when it is no longer needed).
 
-        If the sever is not stopped by the caller and execution reaches the end, the
+        If the server is not stopped by the caller and execution reaches the end, the
         program needs to be terminated by Ctrl+C or by signal as it will not terminate until
         the thread is stopped.
 
-        If the sever is already running :py:class:`HTTPServerError` will be raised. If you are
+        If the server is already running :py:class:`HTTPServerError` will be raised. If you are
         unsure, call :py:meth:`is_running` first.
 
         There's a context interface of this class which stops the server when the context block ends.
@@ -734,7 +759,16 @@ class HTTPServerBase(abc.ABC):  # pylint: disable=too-many-instance-attributes
         if self.is_running():
             raise HTTPServerError("Server is already running")
 
-        self.server = make_server(self.host, self.port, self.application, ssl_context=self.ssl_context)
+        app = Request.application(self.application)
+
+        self.server = make_server(
+            self.host,
+            self.port,
+            app,
+            ssl_context=self.ssl_context,
+            threaded=self.threaded,
+        )
+
         self.port = self.server.port  # Update port (needed if `port` was set to 0)
         self.server_thread = threading.Thread(target=self.thread_target)
         self.server_thread.start()
@@ -749,6 +783,8 @@ class HTTPServerBase(abc.ABC):  # pylint: disable=too-many-instance-attributes
         Only a running server can be stopped. If the sever is not running, :py:class`HTTPServerError`
         will be raised.
         """
+        assert self.server is not None
+        assert self.server_thread is not None
         if not self.is_running():
             raise HTTPServerError("Server is not running")
         self.server.shutdown()
@@ -756,7 +792,7 @@ class HTTPServerBase(abc.ABC):  # pylint: disable=too-many-instance-attributes
         self.server = None
         self.server_thread = None
 
-    def add_assertion(self, obj):
+    def add_assertion(self, obj: str | AssertionError):
         """
         Add a new assertion
 
@@ -812,9 +848,9 @@ class HTTPServerBase(abc.ABC):  # pylint: disable=too-many-instance-attributes
         As the result, there's an assertion added (which can be raised by :py:meth:`check_assertions`).
 
         """
-        text = "No handler found for request {!r}.\n".format(request)
+        text = "No handler found for request {!r} with data {!r}.".format(request, request.data)
         self.add_assertion(text + extra_message)
-        return Response("No handler found for this request", self.no_handler_status_code)
+        return Response(text + extra_message, self.no_handler_status_code)
 
     @abc.abstractmethod
     def dispatch(self, request: Request) -> Response:
@@ -824,10 +860,8 @@ class HTTPServerBase(abc.ABC):  # pylint: disable=too-many-instance-attributes
         :param request: the request object from the werkzeug library
         :return: the response object what the handler responded, or a response which contains the error
         """
-        pass
 
-    @Request.application  # type: ignore
-    def application(self, request: Request):
+    def application(self, request: Request) -> Response:
         """
         Entry point of werkzeug.
 
@@ -853,7 +887,12 @@ class HTTPServerBase(abc.ABC):  # pylint: disable=too-many-instance-attributes
             self.start()
         return self
 
-    def __exit__(self, *args, **kwargs):
+    def __exit__(
+        self,
+        exc_type: type[BaseException] | None,
+        exc_value: BaseException | None,
+        traceback: TracebackType | None,
+    ):
         """
         Provide the context API
 
@@ -864,7 +903,7 @@ class HTTPServerBase(abc.ABC):  # pylint: disable=too-many-instance-attributes
             self.stop()
 
     @staticmethod
-    def format_host(host):
+    def format_host(host: str):
         """
         Formats a hostname so it can be used in a URL.
         Notably, this adds brackets around IPV6 addresses when
@@ -893,6 +932,8 @@ class HTTPServer(HTTPServerBase):  # pylint: disable=too-many-instance-attribute
     :param default_waiting_settings: the waiting settings object to use as default settings for :py:meth:`wait` context
         manager
 
+    :param threaded: whether to handle concurrent requests in separate threads
+
     .. py:attribute:: no_handler_status_code
 
         Attribute containing the http status code (int) which will be the response
@@ -905,17 +946,19 @@ class HTTPServer(HTTPServerBase):  # pylint: disable=too-many-instance-attribute
 
     def __init__(
         self,
-        host=DEFAULT_LISTEN_HOST,
-        port=DEFAULT_LISTEN_PORT,
-        ssl_context: Optional[SSLContext] = None,
-        default_waiting_settings: Optional[WaitingSettings] = None,
+        host: str = DEFAULT_LISTEN_HOST,
+        port: int = DEFAULT_LISTEN_PORT,
+        ssl_context: SSLContext | None = None,
+        default_waiting_settings: WaitingSettings | None = None,
+        *,
+        threaded: bool = False,
     ):
         """
         Initializes the instance.
         """
-        super().__init__(host, port, ssl_context)
+        super().__init__(host, port, ssl_context, threaded=threaded)
 
-        self.ordered_handlers: List[RequestHandler] = []
+        self.ordered_handlers: list[RequestHandler] = []
         self.oneshot_handlers = RequestHandlerList()
         self.handlers = RequestHandlerList()
         self.permanently_failed = False
@@ -946,15 +989,31 @@ class HTTPServer(HTTPServerBase):  # pylint: disable=too-many-instance-attribute
         self.oneshot_handlers = RequestHandlerList()
         self.handlers = RequestHandlerList()
 
+    def expect(self, matcher: RequestMatcher, handler_type: HandlerType = HandlerType.PERMANENT) -> RequestHandler:
+        """
+        Create and register a request handler.
+
+        :param matcher: :py:class:`RequestMatcher` used to match requests.
+        :param handler_type: type of handler
+        """
+        request_handler = RequestHandler(matcher)
+        if handler_type == HandlerType.PERMANENT:
+            self.handlers.append(request_handler)
+        elif handler_type == HandlerType.ONESHOT:
+            self.oneshot_handlers.append(request_handler)
+        elif handler_type == HandlerType.ORDERED:
+            self.ordered_handlers.append(request_handler)
+        return request_handler
+
     def expect_request(
         self,
-        uri: Union[str, URIPattern, Pattern[str]],
+        uri: str | URIPattern | Pattern[str],
         method: str = METHOD_ALL,
-        data: Union[str, bytes, None] = None,
+        data: str | bytes | None = None,
         data_encoding: str = "utf-8",
-        headers: Optional[Mapping[str, str]] = None,
-        query_string: Union[None, QueryMatcher, str, bytes, Mapping] = None,
-        header_value_matcher: Optional[HVMATCHER_T] = None,
+        headers: Mapping[str, str] | None = None,
+        query_string: None | QueryMatcher | str | bytes | Mapping[str, str] = None,
+        header_value_matcher: HVMATCHER_T | None = None,
         handler_type: HandlerType = HandlerType.PERMANENT,
         json: Any = UNDEFINED,
     ) -> RequestHandler:
@@ -971,7 +1030,15 @@ class HTTPServer(HTTPServerBase):  # pylint: disable=too-many-instance-attribute
         If `handler_type` is `HandlerType.ORDERED` an ordered request handler is created. Comparing to oneshot handler,
         ordered handler also determines the order of the requests to be served. For example if there are two ordered
         handlers registered, the first request must hit the first handler, and the second request must hit the second
-        one, and not vice versa. If one or more ordered handler defined, those must be exhausted first.
+        one, and not vice versa. If one or more ordered handler defined, those
+        must be exhausted first.
+
+        .. note::
+
+            Once this method is called, the response should also be specified by
+            calling one of the respond methods of the returned
+            :py:class:`RequestHandler` object, otherwise
+            :py:class:`NoHandlerError` will be raised on an incoming request.
 
 
         :param uri: URI of the request. This must be an absolute path starting with ``/``, a
@@ -1023,13 +1090,13 @@ class HTTPServer(HTTPServerBase):  # pylint: disable=too-many-instance-attribute
 
     def expect_oneshot_request(
         self,
-        uri: Union[str, URIPattern, Pattern[str]],
+        uri: str | URIPattern | Pattern[str],
         method: str = METHOD_ALL,
-        data: Union[str, bytes, None] = None,
+        data: str | bytes | None = None,
         data_encoding: str = "utf-8",
-        headers: Optional[Mapping[str, str]] = None,
-        query_string: Union[None, QueryMatcher, str, bytes, Mapping] = None,
-        header_value_matcher: Optional[HVMATCHER_T] = None,
+        headers: Mapping[str, str] | None = None,
+        query_string: None | QueryMatcher | str | bytes | Mapping[str, str] = None,
+        header_value_matcher: HVMATCHER_T | None = None,
         json: Any = UNDEFINED,
     ) -> RequestHandler:
         """
@@ -1078,13 +1145,13 @@ class HTTPServer(HTTPServerBase):  # pylint: disable=too-many-instance-attribute
 
     def expect_ordered_request(
         self,
-        uri: Union[str, URIPattern, Pattern[str]],
+        uri: str | URIPattern | Pattern[str],
         method: str = METHOD_ALL,
-        data: Union[str, bytes, None] = None,
+        data: str | bytes | None = None,
         data_encoding: str = "utf-8",
-        headers: Optional[Mapping[str, str]] = None,
-        query_string: Union[None, QueryMatcher, str, bytes, Mapping] = None,
-        header_value_matcher: Optional[HVMATCHER_T] = None,
+        headers: Mapping[str, str] | None = None,
+        query_string: None | QueryMatcher | str | bytes | Mapping[str, str] = None,
+        header_value_matcher: HVMATCHER_T | None = None,
         json: Any = UNDEFINED,
     ) -> RequestHandler:
         """
@@ -1141,13 +1208,13 @@ class HTTPServer(HTTPServerBase):  # pylint: disable=too-many-instance-attribute
         This method is primarily used when reporting errors.
         """
 
-        def format_handlers(handlers):
+        def format_handlers(handlers: list[RequestHandler]):
             if handlers:
                 return ["    {!r}".format(handler.matcher) for handler in handlers]
             else:
                 return ["    none"]
 
-        lines = []
+        lines: list[str] = []
         lines.append("Ordered matchers:")
         lines.extend(format_handlers(self.ordered_handlers))
         lines.append("")
@@ -1168,7 +1235,7 @@ class HTTPServer(HTTPServerBase):  # pylint: disable=too-many-instance-attribute
 
         """
         if self._waiting_settings.stop_on_nohandler:
-            self._set_waiting_result(False)
+            self._set_waiting_result(value=False)
 
         return super().respond_nohandler(request, self.format_matchers() + extra_message)
 
@@ -1246,7 +1313,7 @@ class HTTPServer(HTTPServerBase):  # pylint: disable=too-many-instance-attribute
 
         return response
 
-    def _set_waiting_result(self, value: bool) -> None:
+    def _set_waiting_result(self, value: bool) -> None:  # noqa: FBT001
         """Set waiting_result
 
         Setting is implemented as putting value to queue without waiting. If queue is full we simply ignore the
@@ -1257,14 +1324,14 @@ class HTTPServer(HTTPServerBase):  # pylint: disable=too-many-instance-attribute
 
     def _update_waiting_result(self) -> None:
         if not self.oneshot_handlers and not self.ordered_handlers:
-            self._set_waiting_result(True)
+            self._set_waiting_result(value=True)
 
     @contextmanager
     def wait(
         self,
-        raise_assertions: Optional[bool] = None,
-        stop_on_nohandler: Optional[bool] = None,
-        timeout: Optional[float] = None,
+        raise_assertions: bool | None = None,
+        stop_on_nohandler: bool | None = None,
+        timeout: float | None = None,
     ):
         """Context manager to wait until the first of following event occurs: all ordered and oneshot handlers were
         executed, unexpected request was received (if `stop_on_nohandler` is set to `True`), or time was out
@@ -1314,7 +1381,77 @@ class HTTPServer(HTTPServerBase):  # pylint: disable=too-many-instance-attribute
             waiting.complete(result=False)
             if self._waiting_settings.raise_assertions:
                 raise AssertionError(
-                    "Wait timeout occurred, but some handlers left:\n" "{}".format(self.format_matchers())
+                    "Wait timeout occurred, but some handlers left:\n{}".format(self.format_matchers())
                 )
         if self._waiting_settings.raise_assertions and not waiting.result:
             self.check_assertions()
+
+    def iter_matching_requests(self, matcher: RequestMatcher) -> Iterable[tuple[Request, Response]]:
+        """
+        Queries log for matching requests.
+
+
+        :param matcher: the matcher object to match requests
+        :return: an iterator with request-response pair from the log
+        """
+
+        for request, response in self.log:
+            if matcher.match(request):
+                yield (request, response)
+
+    def get_matching_requests_count(self, matcher: RequestMatcher) -> int:
+        """
+        Queries the log for matching requests, returning the number of log
+        entries matching for the specified matcher.
+
+        :param matcher: the matcher object to match requests
+        :return: the number of log entries matching
+        """
+        return len(list(self.iter_matching_requests(matcher)))
+
+    def assert_request_made(self, matcher: RequestMatcher, *, count: int = 1):
+        """
+        Check the amount of log entries matching for the matcher specified. By
+        default it verifies that exactly one request matching for the matcher
+        specified. The expected count can be customized with the count kwarg
+        (including zero, which asserts that no requests made for the given
+        matcher).
+
+        :param matcher: the matcher object to match requests
+        :param count: the expected number of matches in the log
+        :return: ``None`` if the assert succeeded, raises
+            :py:class:`AssertionError` if not.
+        """
+
+        matching_count = self.get_matching_requests_count(matcher)
+        if matching_count != count:
+            similar_requests: list[Request] = []
+            for request, _ in self.log:
+                if request.path == matcher.uri:
+                    similar_requests.append(request)
+
+            assert_msg_lines = [
+                f"Matching request found {matching_count} times but expected {count} times.",
+                f"Expected request: {matcher}",
+            ]
+
+            if similar_requests:
+                assert_msg_lines.append(f"Found {len(similar_requests)} similar request(s):")
+                for request in similar_requests:
+                    assert_msg_lines.extend(
+                        (
+                            "--- Similar Request Start",
+                            f"Path: {request.path}",
+                            f"Method: {request.method}",
+                            f"Body: {request.get_data()!r}",
+                            f"Headers: {request.headers}",
+                            f"Query String: {request.query_string.decode('utf-8')!r}",
+                            "--- Similar Request End",
+                        )
+                    )
+            else:
+                assert_msg_lines.append("No similar requests found.")
+
+            assert_msg = "\n".join(assert_msg_lines) + "\n"
+
+            assert matching_count == count, assert_msg
